"""
Zelt: Orchestrate Locust deployments in Kubernetes.

Usage:
    zelt from-har <har-files>... -m <manifests>
                                 [-w <pods>]
                                 [--storage <method>]
                                 [--s3-bucket <name> --s3-key <name>]
                                 [-p <plugin-name>]...
                                 [--clean]
                                 [--logging <level>]
    zelt from-har <har-files>... --local
                                 [-p <plugin-name>]...
                                 [--logging <level>]
    zelt from-har --config <file>
                  [--local]
                  [--clean]
                  [--logging <level>]
    zelt from-locustfile <locustfile> -m <manifests>
                                      [-w <pods>]
                                      [--storage <method>]
                                      [--s3-bucket <name> --s3-key <name>]
                                      [--clean]
                                      [--logging <level>]
    zelt from-locustfile <locustfile> --local
                                      [--logging <level>]
    zelt from-locustfile --config <file>
                         [--local]
                         [--clean]
                         [--logging <level>]
    zelt rescale <required-pods> -m <manifests>
                                 [--logging <level>]
    zelt rescale <required-pods> --config <file>
                                 [--logging <level>]
    zelt delete -m <manifests> [--storage <method>]
                               [--s3-bucket <name> --s3-key <name>]
                               [--logging <level>]
    zelt delete --config <file>
                [--logging <level>]
    zelt --help
    zelt --version

Options:
    -h, --help                               Show this screen.
    -v, --version                            Show version.
    -p, --transformer-plugins=<plugin-name>  Module name of Transformer plugin (repeatable).
    -m, --manifests=<manifests>              Path to manifest files.
    -w, --worker-pods=<pods>                 Number of worker pods to deploy [default: 1].
    -s, --storage=<method>                   Remote locustfile storage method (S3 or ConfigMap)
                                               [default: ConfigMap].
    --s3-bucket=<name>                       Name of S3 bucket for remote locustfile storage.
    --s3-key=<name>                          Name of S3 key for remote locustfile storage.
    -c, --clean                              Delete and redeploy remote resources.
    -l, --local                              Run Locust locally.
    --logging=<level>                        Set logging level (INFO, DEBUG, or ERROR) [default: INFO].
    --config=<file>                          Optional configuration file specifying options.
"""


import logging
import os
import pkg_resources
import sys
import yaml
from docopt import docopt
from pathlib import Path
from typing import NamedTuple, Sequence

import zelt
from zelt.zelt import StorageMethod


class Config(NamedTuple):

    from_har: bool
    from_locustfile: bool
    rescale: bool
    delete: bool
    har_files: Sequence[os.PathLike]
    locustfile: os.PathLike
    transformer_plugins: Sequence[str]
    manifests: os.PathLike
    worker_pods: int
    required_pods: int
    storage: str
    s3_bucket: str
    s3_key: str
    clean: bool
    local: bool
    logging: str


def cli():
    """
    Entrypoint for Zelt.
    """

    # Disable deprecation warning coming from Kubernetes client's YAML loading.
    # See https://github.com/yaml/pyyaml/wiki/PyYAML-yaml.load(input)-Deprecation
    yaml.warnings({"YAMLLoadWarning": False})

    config = _load_config(docopt(__doc__, version=_version()))

    logging.basicConfig(level=config.logging)

    if config.from_har:
        config = config._replace(
            locustfile=zelt.invoke_transformer(
                paths=config.har_files, plugin_names=config.transformer_plugins
            )
        )
        _deploy(config)

    if config.from_locustfile:
        _deploy(config)

    if config.rescale:
        _rescale(config)

    if config.delete:
        _delete(config)


def _version() -> str:
    return pkg_resources.get_distribution("zelt").version


def _deploy(config: Config) -> None:
    """
    Deploys Locust.
    """
    try:
        zelt.deploy(
            config.locustfile,
            int(config.worker_pods),
            config.manifests,
            config.clean,
            StorageMethod.from_storage_arg(config.storage),
            config.local,
            config.s3_bucket,
            config.s3_key,
        )
    except Exception as e:
        logging.fatal("Error: %s", e)
        exit(1)


def _rescale(config: Config) -> None:
    """
    Rescales a worker deployment.
    """
    try:
        zelt.rescale(config.manifests, int(config.required_pods))
    except Exception as e:
        logging.fatal("Error: %s", e)
        exit(1)


def _delete(config: Config) -> None:
    """
    Deletes a deployment.
    """
    try:
        zelt.delete(
            config.manifests,
            StorageMethod.from_storage_arg(config.storage),
            config.s3_bucket,
            config.s3_key,
        )
    except Exception as e:
        logging.fatal("Error: %s", e)
        exit(1)


def _load_config(config: dict) -> Config:
    """
    Loads config from command-line or file.
    """
    config = _normalise_config(config)

    if config["config"]:
        config = {**config, **yaml.safe_load(Path(config["config"]).read_text())}

    return Config(
        from_har=config["from-har"],
        from_locustfile=config["from-locustfile"],
        rescale=config["rescale"],
        delete=config["delete"],
        har_files=config.get("har-files", []),
        locustfile=config["locustfile"],
        transformer_plugins=config.get("transformer-plugins", []),
        manifests=config["manifests"],
        worker_pods=config["worker-pods"],
        required_pods=config["required-pods"],
        storage=config["storage"],
        s3_bucket=config["s3-bucket"],
        s3_key=config["s3-key"],
        clean=config["clean"],
        local=config["local"],
        logging=config["logging"],
    )


def _normalise_config(config: dict) -> dict:
    """
    Removes special characters from config keys.
    """
    normalised_config = {}
    for k in config:
        normalised_config[
            k.replace("--", "").replace("<", "").replace(">", "")
        ] = config[k]
    return normalised_config



import logging
import os

try:
    import boto3
except ImportError as err:
    raise ImportError(
        "boto3 not found. It is required for uploading to S3 with the "
        "'--storage=s3' option. "
        "It can be installed with 'pip install boto3'."
    ) from err

from zelt.kubernetes.storage.protocol import LocustfileStorage


class S3Storage(LocustfileStorage):
    def __init__(self, bucket: str, key: str) -> None:
        super().__init__()
        self.bucket = bucket
        self.key = key

    def upload(self, locustfile: os.PathLike) -> None:
        def _upload_callback(nb_bytes_transferred: int) -> None:
            logging.info(
                "Uploading %s to %s as %s: %s bytes transferred.",
                locustfile,
                self.bucket,
                self.key,
                nb_bytes_transferred,
            )

        boto3.resource("s3").Object(self.bucket, self.key).upload_file(
            Filename=os.fspath(locustfile), Callback=_upload_callback
        )

    def delete(self) -> None:
        logging.info("Deleting %s from S3 bucket %s...", self.key, self.bucket)
        boto3.resource("s3").Object(self.bucket, self.key).delete()



import logging
import os
from pathlib import Path

from kubernetes.client import V1ConfigMap, CoreV1Api, V1DeleteOptions
from kubernetes.client.rest import ApiException

import zelt.kubernetes.client as client
from zelt.kubernetes.storage.protocol import LocustfileStorage

CONFIGMAP_NAME = "zelt-locustfile"
CONFIGMAP_KEY = "locustfile.py"


class ConfigmapStorage(LocustfileStorage):
    def __init__(self, namespace: str, labels: dict) -> None:
        super().__init__()
        self.namespace = namespace
        self.labels = dict(labels)

    def upload(self, locustfile: os.PathLike) -> None:
        logging.info("Creating ConfigMap %r...", CONFIGMAP_NAME)
        config_map = V1ConfigMap(
            data={CONFIGMAP_KEY: Path(locustfile).read_text()},
            metadata={"name": CONFIGMAP_NAME, "labels": self.labels},
        )
        try:
            logging.debug("Creating ConfigMap %r...", CONFIGMAP_NAME)
            CoreV1Api().create_namespaced_config_map(
                namespace=self.namespace, body=config_map
            )
            logging.debug("ConfigMap %r created.", CONFIGMAP_NAME)
        except ApiException as err:
            logging.error(
                "Failed to create ConfigMap %r: %s", CONFIGMAP_NAME, err.reason
            )
            raise

    def delete(self) -> None:
        try:
            logging.info("Deleting ConfigMap %r...", CONFIGMAP_NAME)
            CoreV1Api().delete_namespaced_config_map(
                name=CONFIGMAP_NAME,
                namespace=self.namespace,
                body=V1DeleteOptions(propagation_policy="Foreground"),
            )
            logging.debug("Waiting for ConfigMap %r to be deleted...", CONFIGMAP_NAME)
            client.await_no_resources_found(
                CoreV1Api().list_namespaced_config_map, namespace=self.namespace
            )
            logging.debug("ConfigMap %r deleted.", CONFIGMAP_NAME)
        except ApiException as err:
            if err.status == 404:
                logging.debug(
                    "Skipping ConfigMap %r deletion: %s", CONFIGMAP_NAME, err.reason
                )
                return
            logging.error(
                "Failed to delete ConfigMap %r: %s", CONFIGMAP_NAME, err.reason
            )
            raise



import logging
import os
from tenacity import RetryError

import zelt.kubernetes.client as kube
from zelt.kubernetes.manifest_set import ManifestSet
from zelt.kubernetes.storage.protocol import LocustfileStorage


def create_resources(
    ms: ManifestSet, storage: LocustfileStorage, locustfile: os.PathLike
) -> None:
    try:
        kube.read_config()
        kube.create_namespace(ms.namespace)

        storage.upload(locustfile)

        kube.create_deployment(ms.controller)
        kube.wait_until_pod_ready(ms.controller)

        if ms.worker:
            kube.create_deployment(ms.worker)

        kube.create_service(ms.service)
        kube.create_ingress(ms.ingress)

        if ms.others:
            kube.try_creating_custom_objects(ms.others)

    except (kube.ApiException, RetryError) as err:
        logging.error("Kubernetes operation failed: %s", err.reason)


def delete_resources(ms: ManifestSet, storage: LocustfileStorage) -> None:
    logging.info("Deleting resources...")
    try:
        kube.read_config()
        namespace = ms.namespace.name

        storage.delete()
        kube.delete_ingress(ms.ingress.name, namespace)
        kube.delete_service(ms.service.name, namespace)
        kube.delete_deployments(namespace)
        kube.delete_namespace(namespace)
    except (kube.ApiException, RetryError) as err:
        logging.error("Kubernetes operation failed: %s", err.reason)


def update_worker_pods(ms: ManifestSet, worker_replicas: int) -> None:
    if ms.worker is not None:
        ms.worker.body["spec"]["replicas"] = worker_replicas


def rescale_worker_deployment(ms: ManifestSet, replicas: int) -> None:
    if not ms.worker:
        logging.error(
            "Missing worker manifest. Only worker deployments can be rescaled."
        )
        return

    try:
        kube.read_config()
        kube.rescale_deployment(ms.worker, replicas)
    except kube.ApiException as err:
        logging.error("Kubernetes operation failed: %s", err.reason)



from collections import defaultdict
from os import PathLike
from typing import NamedTuple, List, Dict, Optional

from zelt.kubernetes.manifest import Manifest, ResourceType, DeploymentRole


class ManifestSet(NamedTuple):
    """
    Result of categorizing the user-provided Kubernetes manifests.
    """

    namespace: Manifest
    service: Manifest
    ingress: Manifest
    controller: Manifest
    worker: Optional[Manifest]
    others: List[Manifest]


# TODO: Refactor.
def from_directory(dir_path: PathLike) -> ManifestSet:
    categories: Dict[ResourceType, List[Manifest]] = defaultdict(list)
    for m in Manifest.all_from_directory(dir_path):
        categories[m.kind].append(m)

    # Sanity checks for manifests that must be given exactly once.
    unique_resources = (
        ResourceType.NAMESPACE,
        ResourceType.SERVICE,
        ResourceType.INGRESS,
    )
    for kind in unique_resources:
        wanted = f"resource of kind {kind.value!r}"
        try:
            manifests = categories[kind]
        except KeyError:
            raise ValueError(f"Missing required {wanted}.") from None
        if len(manifests) != 1:
            raise ValueError(f"Expected exactly one {wanted} but got {len(manifests)}.")

    # Sanity checks for deployment manifests.
    deploy_rk = f"resource of kind {ResourceType.DEPLOYMENT.value!r}"
    try:
        deployments = categories[ResourceType.DEPLOYMENT]
    except KeyError:
        raise ValueError(f"Missing required {deploy_rk}.") from None
    if len(deployments) > 1:
        actual_roles = {d.role for d in deployments}
        expected_roles = {DeploymentRole.CONTROLLER, DeploymentRole.WORKER}
        if not expected_roles.issubset(actual_roles):
            raise ValueError(
                "Distributed Locust deployments must have roles "
                f"covering {[r.value for r in expected_roles]}, "
                f"got only {[r.value for r in actual_roles]}."
            )

    controllers = [d for d in deployments if d.role is DeploymentRole.CONTROLLER]
    if len(controllers) != 1:
        raise ValueError(
            "Expected exactly one deployment with role "
            f"{DeploymentRole.CONTROLLER.value!r}."
        )

    workers = [d for d in deployments if d.role is DeploymentRole.WORKER]
    if len(workers) > 1:
        raise ValueError(
            f"Expected at most one deployment with role {DeploymentRole.WORKER.value!r}."
        )

    return ManifestSet(
        namespace=categories[ResourceType.NAMESPACE][0],
        service=categories[ResourceType.SERVICE][0],
        ingress=categories[ResourceType.INGRESS][0],
        controller=controllers[0],
        worker=workers[0] if workers else None,
        others=categories[ResourceType.OTHER],
    )



import enum
import logging
import os
import subprocess
from pathlib import Path
from time import time
from typing import Optional, Sequence
from zlib import adler32

from zelt.kubernetes import deployer, manifest_set
from zelt.kubernetes.manifest_set import ManifestSet
from zelt.kubernetes.storage.configmap import ConfigmapStorage
from zelt.kubernetes.storage.protocol import LocustfileStorage
from zelt.kubernetes.storage.s3 import S3Storage

try:
    import transformer

    TRANSFORMER_NOT_FOUND = False
except ImportError:
    TRANSFORMER_NOT_FOUND = True


class HARFilesNotFoundException(Exception):
    pass


class StorageMethod(enum.Enum):
    CONFIGMAP = enum.auto()
    S3 = enum.auto()

    @classmethod
    def from_storage_arg(cls, arg: str) -> "StorageMethod":
        if arg.lower() == "s3":
            return cls.S3
        if arg.lower() in ("cm", "configmap"):
            return cls.CONFIGMAP
        raise ValueError(f"unknown {cls.__qualname__} {arg!r}")

    def build_storage(
        self,
        manifests: ManifestSet,
        s3_bucket: Optional[str] = None,
        s3_key: Optional[str] = None,
    ) -> LocustfileStorage:
        if self is StorageMethod.S3 and (not (s3_bucket and s3_key)):
            raise ValueError(
                "Missing required 's3-bucket' and/or 's3-key' options "
                "for 'storage=s3' option."
            )

        if self is not StorageMethod.S3 and (s3_bucket or s3_key):
            raise ValueError(
                "Unexpected 's3-bucket' or 's3-key' options "
                "without 'storage=s3' option."
            )

        if self is StorageMethod.S3:
            return S3Storage(bucket=s3_bucket, key=s3_key)

        return ConfigmapStorage(
            namespace=manifests.namespace.name, labels=manifests.namespace.labels_dict
        )


def deploy(
    locustfile: os.PathLike,
    worker_pods: int,
    manifests_path: Optional[os.PathLike],
    clean: bool,
    storage_method: StorageMethod,
    local: bool,
    s3_bucket: Optional[str] = None,
    s3_key: Optional[str] = None,
) -> None:
    if local:
        if manifests_path:
            logging.warning(
                "Mutually incompatible options 'local' and 'manifests' specified. Defaulting to running locally."
            )
        return _deploy_locally(locustfile)

    if not manifests_path:
        raise ValueError("Missing required 'manifests' option.")

    _deploy_in_kubernetes(
        locustfile,
        worker_pods,
        manifests_path=manifests_path,
        clean_deployment=clean,
        storage_method=storage_method,
        s3_bucket=s3_bucket,
        s3_key=s3_key,
    )


def rescale(manifests_path, worker_pods: int) -> None:
    if not manifests_path:
        raise ValueError("Missing required 'manifests' option.")

    if worker_pods < 0:
        raise ValueError(f"Expected a positive number of pods, got {worker_pods}.")

    manifests = manifest_set.from_directory(manifests_path)
    deployer.update_worker_pods(manifests, worker_pods)
    deployer.rescale_worker_deployment(manifests, worker_pods)
    logging.info("Rescaling complete.")


def delete(
    manifests_path: os.PathLike,
    storage_method: StorageMethod,
    s3_bucket: Optional[str] = None,
    s3_key: Optional[str] = None,
) -> None:
    if not manifests_path:
        raise ValueError("Missing required 'manifests' option.")

    manifests = manifest_set.from_directory(manifests_path)
    storage = storage_method.build_storage(manifests, s3_bucket, s3_key)
    deployer.delete_resources(manifests, storage)
    logging.info("Deletion complete.")


def invoke_transformer(
    paths: Sequence[os.PathLike], plugin_names: Sequence[str]
) -> Path:
    if TRANSFORMER_NOT_FOUND:
        raise ImportError(
            "Transformer not found. It is required for calls to 'from-har'. "
            "It can be installed with 'pip install har-transformer'."
        )

    har_files = []
    for path in paths:
        if os.path.exists(path):
            har_files.append(Path(path))
    if not har_files:
        raise HARFilesNotFoundException(f"Could not load any HAR files from {paths}")

    locustfile = Path(f"locustfile-{adler32(str(paths).encode(encoding='utf-8'))}.py")
    with locustfile.open("w") as f:
        transformer.dump(f, har_files, plugin_names)
    logging.info("%s created from %s.", locustfile, har_files)
    return locustfile


def _deploy_locally(locustfile: os.PathLike) -> None:
    logging.info("Deploying Locust locally with locustfile %s...", locustfile)
    logging.info("\n\nOpen http://localhost:8089/ to access the Locust dashboard.\n\n")

    # The host value is unused when full URLs are used in the locustfile.
    subprocess.run(["locust", "-f", os.fspath(locustfile), "--host=unused"], check=True)


def _deploy_in_kubernetes(
    locustfile: os.PathLike,
    worker_pods: int,
    manifests_path: os.PathLike,
    clean_deployment: bool,
    storage_method: StorageMethod,
    s3_bucket: Optional[str],
    s3_key: Optional[str],
) -> None:
    if worker_pods < 0:
        raise ValueError(f"Expected a positive number of pods, got {worker_pods}.")

    logging.info(
        "Deploying Locust in Kubernetes with locustfile %s and %s worker pods...",
        locustfile,
        worker_pods,
    )

    manifests = manifest_set.from_directory(manifests_path)

    storage = storage_method.build_storage(manifests, s3_bucket, s3_key)

    if clean_deployment:
        deployer.delete_resources(manifests, storage)

    deployer.update_worker_pods(manifests, worker_pods)
    deployer.create_resources(manifests, storage, locustfile)

    logging.info(
        "\n\nOpen %s to access the Locust dashboard.\n\n", manifests.ingress.host
    )



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['zelt', 'zelt.kubernetes', 'zelt.kubernetes.storage']

package_data = \
{'': ['*']}

modules = \
['main']
install_requires = \
['boto3>=1.9,<2.0',
 'docopt>=0.6.2,<0.7.0',
 'greenlet>=0.4.15,<0.5.0',
 'har-transformer>=1.0,<2.0',
 'kubernetes>=10.0.1,<11.0.0',
 'locustio>=0.9.0,<0.10.0',
 'pyyaml>=5.1,<6.0',
 'tenacity>=5.0,<6.0']

extras_require = \
{'docs': ['sphinx>=1.8,<2.0',
          'sphinx-autodoc-typehints>=1.6,<2.0',
          'sphinx-issues>=1.2,<2.0']}

entry_points = \
{'console_scripts': ['zelt = main:cli']}

setup_kwargs = {
    'name': 'zelt',
    'version': '1.2.14',
    'description': 'Zalando end-to-end load tester',
    'long_description': ".. image:: docs/_static/zelt.png\n    :alt: Zelt logo\n\nZalando end-to-end load tester\n******************************\n\n.. image:: https://travis-ci.org/zalando-incubator/zelt.svg?branch=master\n   :alt: travis-ci status badge\n   :target: https://travis-ci.org/zalando-incubator/zelt\n\n.. image:: https://badgen.net/pypi/v/zelt\n   :alt: pypi version badge\n   :target: https://pypi.org/project/zelt\n\n.. image:: https://api.codacy.com/project/badge/Grade/a74dee2bbbd64da8951a3cec5059dda3\n   :alt: code quality badge\n   :target: https://www.codacy.com/app/bmaher/zelt\n\n.. image:: https://api.codacy.com/project/badge/Coverage/a74dee2bbbd64da8951a3cec5059dda3\n   :alt: test coverage badge\n   :target: https://www.codacy.com/app/bmaher/zelt\n\n.. image:: https://badgen.net/badge/code%20style/black/000\n   :alt: Code style: Black\n   :target: https://github.com/ambv/black\n\n|\n\nA **command-line tool** for orchestrating the deployment of\nLocust_ in Kubernetes_.\n\nUse it in conjunction with Transformer_ to run large-scale end-to-end\nload testing of your website.\n\nPrerequistes\n============\n\n-  `Python 3.6+`_\n\nInstallation\n============\n\nInstall using pip:\n\n.. code:: bash\n\n   pip install zelt\n\nUsage\n=====\n\nExample HAR files, locustfile, and manifests are included in the\n``examples/`` directory, try them out.\n\n**N.B** The cluster to deploy to is determined by your currently\nconfigured context. Ensure you are `using the correct cluster`_\nbefore using Zelt.\n\nLocustfile as input\n-------------------\n\nZelt can deploy Locust with a locustfile to a cluster:\n\n.. code:: bash\n\n   zelt from-locustfile PATH_TO_LOCUSTFILE --manifests PATH_TO_MANIFESTS\n\nHAR files(s) as input\n---------------------\n\nZelt can transform HAR file(s) into a locustfile and deploy it along\nwith Locust to a cluster:\n\n.. code:: bash\n\n   zelt from-har PATH_TO_HAR_FILES --manifests PATH_TO_MANIFESTS\n\n**N.B** This requires\nTransformer_ to be installed. For more information about Transformer,\nplease refer to `Transformer's documentation`_.\n\nRescale a deployment\n--------------------\n\nZelt can rescale the number of workers_ in a deployment it has made\nto a cluster:\n\n.. code:: bash\n\n   zelt rescale NUMBER_OF_WORKERS --manifests PATH_TO_MANIFESTS\n\nDelete a deployment\n-------------------\n\nZelt can delete deployments it has made from a cluster:\n\n.. code:: bash\n\n   zelt delete --manifests PATH_TO_MANIFESTS\n\nRun Locust locally\n------------------\n\nZelt can also run Locust locally by providing the ``--local/-l`` flag to\neither the ``from-har`` or ``from-locustfile`` command e.g.:\n\n.. code:: bash\n\n   zelt from-locustfile PATH_TO_LOCUSTFILE --local\n\nUse S3 for locustfile storage\n-----------------------------\n\nBy default, Zelt uses a ConfigMap for storing the locustfile. ConfigMaps\nhave a file-size limitation of ~2MB. If your locustfile is larger than\nthis then you can use an S3 bucket for locustfile storage.\n\nTo do so, add the following parameters to your Zelt command:\n\n-  ``--storage s3``: Switch to S3 storage\n-  ``--s3-bucket``: The name of your S3 bucket\n-  ``--s3-key``: The name of the file as stored in S3\n\n**N.B.** Zelt will *not* create the S3 bucket for you.\n\n**N.B.** Make sure to update your deployment manifest(s) to download the\nlocustfile file from S3 instead of loading from the ConfigMap volume\nmount.\n\nUse a configuration file for Zelt options\n-----------------------------------------\n\nAn alternative to specifying Zelt’s options on the command-line is to\nuse a configuration file, for example:\n\n.. code:: bash\n\n   zelt from-har --config examples/config/config.yaml\n\n**N.B.** The configuration file’s keys are the same as the command-line\noption names but without the double dash (``--``).\n\nDocumentation\n=============\n\nTake a look at our documentation_ for more details.\n\nContributing\n============\n\nPlease read `CONTRIBUTING.md <CONTRIBUTING.md>`__ for details on our\nprocess for submitting pull requests to us, and please ensure you follow\nthe `CODE_OF_CONDUCT.md <CODE_OF_CONDUCT.md>`__.\n\nVersioning\n==========\n\nWe use SemVer_ for versioning.\n\nAuthors\n=======\n\n-  **Brian Maher** - `@bmaher`_\n-  **Oliwia Zaremba** - `@tortila`_\n-  **Thibaut Le Page** - `@thilp`_\n\nSee also the list of `contributors <CONTRIBUTORS.md>`__ who participated\nin this project.\n\nLicense\n=======\n\nThis project is licensed under the MIT License - see the\n`LICENSE <LICENSE>`__ file for details\n\n.. _Locust: https://locust.io/\n.. _Kubernetes: https://kubernetes.io/\n.. _Transformer: https://github.com/zalando-incubator/transformer\n.. _`Python 3.6+`: https://www.python.org/downloads/\n.. _`using the correct cluster`: https://kubernetes.io/docs/reference/kubectl/cheatsheet/#kubectl-context-and-configuration\n.. _`Transformer's documentation`: https://transformer.readthedocs.io/\n.. _workers: https://docs.locust.io/en/stable/running-locust-distributed.html\n.. _documentation: https://zelt.readthedocs.io/\n.. _`@bmaher`: https://github.com/bmaher\n.. _`@tortila`: https://github.com/tortila\n.. _`@thilp`: https://github.com/thilp\n.. _SemVer: http://semver.org/\n",
    'author': 'Brian Maher',
    'author_email': 'brian.maher@zalando.de',
    'url': 'https://github.com/zalando-incubator/zelt',
    'packages': packages,
    'package_data': package_data,
    'py_modules': modules,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import logging
from typing import List, Optional, Callable

from kubernetes import config
from kubernetes.client import (
    CoreV1Api,
    V1Namespace,
    AppsV1Api,
    V1Pod,
    V1Service,
    V1DeleteOptions,
    V1Deployment,
    V1Status,
    V1ContainerStatus,
    NetworkingV1beta1Api,
    NetworkingV1beta1Ingress,
    CustomObjectsApi,
    ApiextensionsV1beta1Api,
)
from kubernetes.client.rest import ApiException
from tenacity import retry, stop_after_delay, wait_fixed, retry_if_exception_type

from .manifest import Manifest

KUBE_API_LIST_TIMEOUT = 360
KUBE_API_DELETE_TIMEOUT = 240
KUBE_API_WAIT = 1
STATUS_NOT_FOUND = 404
DEFAULT_DELETE_OPTIONS = V1DeleteOptions(propagation_policy="Foreground")


class PodNotReadyError(RuntimeError):
    pass


class ResourceStillThereError(RuntimeError):
    pass


def read_config():
    try:
        config.load_kube_config()
    except FileNotFoundError:
        logging.error("Kubernetes config. not found!")
        raise


def create_namespace(namespace: Manifest) -> V1Namespace:
    logging.info("Creating Namespace %r...", namespace.name)
    try:
        return CoreV1Api().create_namespace(body=namespace.body)
    except ApiException as err:
        logging.error("Failed to create Namespace %r: %s", namespace.name, err.reason)
        raise


def delete_namespace(name: str) -> Optional[V1Status]:
    logging.info("Deleting Namespace %r...", name)
    try:
        CoreV1Api().delete_namespace(name=name, body=DEFAULT_DELETE_OPTIONS)
    except ApiException as err:
        if err.status == STATUS_NOT_FOUND:
            logging.debug("Skipping Namespace %r deletion: %s", name, err.reason)
            return
        logging.error("Failed to delete Namespace %r: %s", name, err.reason)
        raise
    await_no_resources_found(CoreV1Api().read_namespace, name=name)


def create_deployment(deployment: Manifest) -> V1Deployment:
    logging.info("Creating Deployment %r...", deployment.name)
    try:
        return AppsV1Api().create_namespaced_deployment(
            namespace=deployment.namespace, body=deployment.body
        )
    except ApiException as err:
        logging.error("Failed to create Deployment %r: %s", deployment.name, err.reason)
        raise


def rescale_deployment(manifest: Manifest, replicas: int) -> V1Deployment:
    logging.info("Rescaling Deployment %r to %s Replicas...", manifest.name, replicas)

    if replicas < 0:
        raise ValueError(f"Expected a positive number of Replicas, got {replicas}")

    logging.debug("Fetching existing Deployment %r...", manifest.name)
    try:
        deployment = AppsV1Api().read_namespaced_deployment(
            name=manifest.name, namespace=manifest.namespace
        )
    except ApiException as err:
        logging.error("Failed to fetch Deployment: %s", err.reason)
        raise

    logging.debug(
        "Changing Deployment Replicas from %s to %s...",
        deployment.spec.replicas,
        replicas,
    )
    deployment.spec.replicas = replicas

    logging.debug("Redeploying Deployment %r...", manifest.name)
    try:
        return AppsV1Api().replace_namespaced_deployment(
            name=manifest.name, namespace=manifest.namespace, body=deployment
        )
    except ApiException as err:
        logging.error("Failed to redeploy Deployment %r: %s", manifest.name, err.reason)
        raise


def delete_deployments(namespace: str) -> Optional[V1Status]:
    logging.info("Deleting Deployments in Namespace %r...", namespace)
    try:
        AppsV1Api().delete_collection_namespaced_deployment(namespace=namespace)
    except ApiException as err:
        if err.status == STATUS_NOT_FOUND:
            logging.debug("Skipping Deployment deletion: %s", err.reason)
            return
        logging.error(
            "Failed to delete Deployments in Namespace %r: %s", namespace, err.reason
        )
        raise
    await_no_resources_found(
        AppsV1Api().list_namespaced_deployment, namespace=namespace
    )


def create_service(service: Manifest) -> V1Service:
    logging.info("Creating Service %r...", service.name)
    try:
        return CoreV1Api().create_namespaced_service(
            namespace=service.namespace, body=service.body
        )
    except ApiException as err:
        logging.error("Failed to create Service %r: %s", service.name, err.reason)
        raise


def delete_service(name: str, namespace: str) -> Optional[V1Status]:
    logging.info("Deleting Service %r...", name)
    try:
        CoreV1Api().delete_namespaced_service(
            name=name, namespace=namespace, body=DEFAULT_DELETE_OPTIONS
        )
    except ApiException as err:
        if err.status == STATUS_NOT_FOUND:
            logging.debug("Skipping Service %r deletion: %s", name, err.reason)
            return
        logging.error("Failed to delete Service %r: %s", name, err.reason)
        raise
    await_no_resources_found(CoreV1Api().list_namespaced_service, namespace=namespace)


def create_ingress(ingress: Manifest) -> NetworkingV1beta1Ingress:
    logging.info("Creating Ingress %r...", ingress.name)
    try:
        return NetworkingV1beta1Api().create_namespaced_ingress(
            namespace=ingress.namespace, body=ingress.body
        )
    except ApiException as err:
        logging.error("Failed to create Ingress %r: %s", ingress.name, err.reason)
        raise


def delete_ingress(name: str, namespace: str) -> Optional[V1Status]:
    logging.info("Deleting Ingress %r...", name)
    try:
        NetworkingV1beta1Api().delete_namespaced_ingress(
            name=name, namespace=namespace, body=DEFAULT_DELETE_OPTIONS
        )
    except ApiException as err:
        if err.status == STATUS_NOT_FOUND:
            logging.debug("Skipping Ingress %r deletion: %s", name, err.reason)
            return
        logging.error("Failed to delete Ingress %r: %s", name, err.reason)
        raise
    await_no_resources_found(
        NetworkingV1beta1Api().list_namespaced_ingress, namespace=namespace
    )


def try_creating_custom_objects(manifests: List[Manifest]):
    logging.info("Fetching CRDs available in the cluster...")
    try:
        custom_resources = (
            ApiextensionsV1beta1Api().list_custom_resource_definition().items
        )
    except ApiException as err:
        logging.error("Failed to fetch CRDs: %s", err.reason)
        raise

    available_kinds = {r.spec.names.kind.lower() for r in custom_resources}

    for m in manifests:
        logging.info("Found a custom manifest: %s %r", m.body["kind"], m.name)
        if m.body["kind"].lower() not in available_kinds:
            logging.error(
                "Unsupported custom manifest %r of kind %r is ignored. "
                "Supported custom resource types are: %s",
                m.name,
                m.body["kind"],
                available_kinds,
            )
            continue

        # By supporting only namespaced resources we don't have to manage
        # the cleanup - it will be handled by the deletion of the namespace.
        matching_resources = [
            r
            for r in custom_resources
            if r.spec.names.kind.lower() == m.body["kind"].lower()
            and r.spec.scope.lower() == "namespaced"
        ]
        if not matching_resources:
            logging.error(
                "Failed to match %r to a namespaced custom resource "
                "definition. Non-namespaced resources are not supported!",
                m.body["kind"],
            )
            continue

        _create_custom_object_with_plural(
            custom_object=m, plural=matching_resources[0].spec.names.plural
        )


def _create_custom_object_with_plural(custom_object: Manifest, plural: str):
    logging.info("Creating %s %r ", custom_object.body["kind"], custom_object.name)
    try:
        group, version = custom_object.body.get("apiVersion").rsplit("/", 1)
        return CustomObjectsApi().create_namespaced_custom_object(
            namespace=custom_object.namespace,
            body=custom_object.body,
            group=group,
            version=version,
            plural=plural,
        )
    except ApiException as err:
        logging.error(
            "Failed to create %s %r: %s",
            custom_object.body["kind"],
            custom_object.name,
            err.reason,
        )
        raise


@retry(
    stop=stop_after_delay(KUBE_API_DELETE_TIMEOUT),
    wait=wait_fixed(KUBE_API_WAIT),
    retry=retry_if_exception_type(ResourceStillThereError),
)
def await_no_resources_found(list_resources: Callable, **kwargs):
    try:
        found = list_resources(**kwargs)
    except ApiException as err:
        if err.status == STATUS_NOT_FOUND:
            return
        raise
    if hasattr(found, "items"):
        found = found.items
    if found:
        raise ResourceStillThereError(f"Resource(s): {found} still found; retrying.")


@retry(
    stop=stop_after_delay(KUBE_API_LIST_TIMEOUT),
    wait=wait_fixed(KUBE_API_WAIT),
    retry=retry_if_exception_type(PodNotReadyError),
)
def wait_until_pod_ready(deployment: Manifest) -> None:
    pod_ready = _pod_status(deployment).ready
    if not pod_ready:
        raise PodNotReadyError()


@retry(stop=stop_after_delay(KUBE_API_LIST_TIMEOUT), wait=wait_fixed(KUBE_API_WAIT))
def _pod_status(deployment: Manifest) -> V1ContainerStatus:
    pod_list = _list_pod(deployment.namespace, deployment.labels)
    return pod_list[0].status.container_statuses[0]


def _list_pod(namespace: str, labels: str) -> List[V1Pod]:
    logging.debug("Listing Pod(s) in Namespace %r with Labels %r...", namespace, labels)
    try:
        return (
            CoreV1Api()
            .list_namespaced_pod(namespace=namespace, label_selector=labels)
            .items
        )
    except ApiException as err:
        logging.error(
            "Failed to list Pod(s) in Namespace %r with Labels %r: %s",
            namespace,
            labels,
            err.reason,
        )
        raise



import logging
import os
from enum import Enum
from pathlib import Path
from typing import NamedTuple, List, Dict

import yaml


class ManifestsNotFoundException(Exception):
    pass


class ResourceType(Enum):
    INGRESS = "Ingress"
    SERVICE = "Service"
    NAMESPACE = "Namespace"
    DEPLOYMENT = "Deployment"
    OTHER = ...


class DeploymentRole(Enum):
    CONTROLLER = "controller"
    WORKER = "worker"
    OTHER = ...


class Manifest(NamedTuple):
    body: dict

    def metadata(self, key: str):
        """
        Returns the value of metadata *key*.

        :raise ValueError: If *key* is not a metadata of this manifest.
        """
        try:
            return self.body["metadata"][key]
        except KeyError:
            raise ValueError(f"no metadata {key!r} in manifest") from None

    def nonempty_string_metadata(self, key: str) -> str:
        """
        Returns the non-empty string corresponding to metadata *key*.

        :raise ValueError: If *key* is not a metadata of this manifest, or if
            its value is not a non-empty string.
        """
        v = self.metadata(key)
        _assert_metadata_type(v, str, key)
        if not v:
            raise ValueError(f"unexpectedly empty string for {key!r} metadata")
        return v

    @property
    def kind(self) -> ResourceType:
        try:
            value = self.body["kind"]
        except KeyError:
            raise ValueError("no kind specified in manifest") from None

        try:
            return ResourceType(value.capitalize())
        except ValueError:
            logging.debug(
                "Non-standard resource type %r converted into %s.",
                value,
                ResourceType.OTHER,
            )
            return ResourceType.OTHER

    @property
    def name(self) -> str:
        return self.nonempty_string_metadata("name")

    @property
    def namespace(self) -> str:
        if self.kind is ResourceType.NAMESPACE:
            return self.name
        return self.nonempty_string_metadata("namespace")

    @property
    def labels(self) -> str:
        key = "labels"
        labels = self.metadata(key)
        _assert_metadata_type(labels, dict, key)
        return ",".join(["{}={}".format(k, v) for k, v in labels.items()])

    @property
    def labels_dict(self) -> Dict[str, str]:
        return dict(self.body.get("metadata", {}).get("labels", {}))

    @property
    def role(self) -> DeploymentRole:
        role = self.labels_dict.get("role", "")
        try:
            return DeploymentRole(role.strip().lower())
        except ValueError:
            return DeploymentRole.OTHER

    @property
    def host(self) -> str:
        pretty_key = "spec.rules[0].host"
        try:
            host = self.body["spec"]["rules"][0]["host"]
        except (KeyError, IndexError):
            raise ValueError(f"no {pretty_key} in manifest")
        _assert_type_as(host, str, f"as {pretty_key} in manifest")
        return host

    @classmethod
    def from_file(cls, file: Path) -> "Manifest":
        """
        :raise ValueError: If no manifest can be read from *file*.
        """
        try:
            body = yaml.safe_load(file.read_text())
        except OSError as err:
            raise ValueError(f"can't read manifest from file {file}") from err
        except yaml.YAMLError as err:
            raise ValueError(f"can't read manifest from non-YAML file {file}") from err
        _assert_type_as(body, dict, f"as top-level manifest object in file {file}")
        return Manifest(body=body)

    @classmethod
    def all_from_directory(cls, manifests_path: os.PathLike) -> List["Manifest"]:
        manifests = []
        for path in Path(manifests_path).glob("*"):
            if path.is_file():
                try:
                    manifests.append(Manifest.from_file(path))
                except ValueError as err:
                    logging.warning("Ignoring %s: %s.", path, err)
        if not manifests:
            raise ManifestsNotFoundException(
                f"Could not load any manifest files from {manifests_path}"
            )
        return manifests


def _assert_type_as(value, t: type, as_msg: str) -> None:
    if not isinstance(value, t):
        raise ValueError(f"expected a {t.__qualname__} but got {value!r} {as_msg}")


def _assert_metadata_type(value, t: type, key: str) -> None:
    _assert_type_as(value, t, f"as {key!r} metadata")
